/*
   Copyright (C) 1998-2000  Ulric Eriksson <ulric@siag.nu>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the Licence, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "../common/cmalloc.h"
#include "../common/fonts.h"
#include "xfonts.h"
#include <X11/xpm.h>
#include "embed.h"
#include "Ruler.h"

#include "../common/richchar.h"
#include "../common/traceme.h"

#include "RichtextP.h"

static void plugin_coords(Widget, XtPointer, int *, int *);
static int default_row_height(XtPointer, int);
static int default_adj_horiz(XtPointer, int);
static int default_style(XtPointer, int);
static rich_char *default_text(XtPointer, int);
static Boolean default_bop(XtPointer, int);

static float floatOne = 1.0;

#define offset(field) XtOffsetOf(RichtextRec, richtext.field)
static XtResource resources[] = {
	{
		XtNrichtextTopRow,	/* name */
		XtCRichtextTopRow,	/* class */
		XtRInt,			/* type */
		sizeof(int),		/* size */
		offset(top_row),	/* offset */
		XtRImmediate,		/* default_type */
		(XtPointer)1		/* default_addr */
	}, {
		XtNrichtextTopCol,
		XtCRichtextTopCol,
		XtRInt,
		sizeof(int),
		offset(top_col),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNrichtextSelectTopRow,
		XtCRichtextSelectTopRow,
		XtRInt,
		sizeof(int),
		offset(sel_top_row),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNrichtextSelectTopCol,
		XtCRichtextSelectTopCol,
		XtRInt,
		sizeof(int),
		offset(sel_top_col),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNrichtextSelectBottomRow,
		XtCRichtextSelectBottomRow,
		XtRInt,
		sizeof(int),
		offset(sel_bottom_row),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNrichtextSelectBottomCol,
		XtCRichtextSelectBottomCol,
		XtRInt,
		sizeof(int),
		offset(sel_bottom_col),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNrichtextPointRow,
		XtCRichtextPointRow,
		XtRInt,
		sizeof(int),
		offset(point_row),
		XtRImmediate,
		(XtPointer)1
	}, {
		XtNrichtextPointCol,
		XtCRichtextPointCol,
		XtRInt,
		sizeof(int),
		offset(point_col),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNrichtextRowHeight,
		XtCRichtextRowHeight,
		XtRPointer,
		sizeof(XtPointer),
		offset(row_height),
		XtRImmediate,
		(XtPointer)default_row_height
	}, {
		XtNrichtextAdjHoriz,
		XtCRichtextAdjHoriz,
		XtRPointer,
		sizeof(XtPointer),
		offset(adj_horiz),
		XtRImmediate,
		(XtPointer)default_adj_horiz
	}, {
		XtNrichtextStyle,
		XtCRichtextStyle,
		XtRPointer,
		sizeof(XtPointer),
		offset(style),
		XtRImmediate,
		(XtPointer)default_style
	}, {
		XtNrichtextBop,
		XtCRichtextBop,
		XtRPointer,
		sizeof(XtPointer),
		offset(bop),
		XtRImmediate,
		(XtPointer)default_bop
	}, {
		XtNrichtextText,
		XtCRichtextText,
		XtRPointer,
		sizeof(XtPointer),
		offset(text),
		XtRImmediate,
		(XtPointer)default_text
	}, {
		XtNrichtextData,
		XtCRichtextData,
		XtRPointer,
		sizeof(XtPointer),
		offset(data),
		XtRImmediate,
		(XtPointer) NULL
	}, {
		XtNrichtextRedisplay,
		XtCRichtextRedisplay,
		XtRBoolean,
		sizeof(Boolean),
		offset(redisplay),
		XtRImmediate,
		(XtPointer)False
	}, {
		XtNrichtextVisibleCursor,
		XtCRichtextVisibleCursor,
		XtRBoolean,
		sizeof(Boolean),
		offset(visible_cursor),
		XtRImmediate,
		(XtPointer)True
	}, {
		XtNrichtextPaperWidth,
		XtCRichtextPaperWidth,
		XtRInt,
		sizeof(int),
		offset(paper_width),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNrichtextLeftMargin,
		XtCRichtextLeftMargin,
		XtRInt,
		sizeof(int),
		offset(left_margin),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNrichtextRightMargin,
		XtCRichtextRightMargin,
		XtRInt,
		sizeof(int),
		offset(right_margin),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNrichtextTabDistance,
		XtCRichtextTabDistance,
		XtRInt,
		sizeof(int),
		offset(tab_distance),
		XtRImmediate,
		(XtPointer)36
	}, {
		XtNrichtextPluginCoords,
		XtCRichtextPluginCoords,
		XtRPointer,
		sizeof(XtPointer),
		offset(plugin_coords),
		XtRImmediate,
		(XtPointer)plugin_coords
	}, {
		XtNzoom,
		XtCZoom,
		XtRFloat,
		sizeof(float),
		offset(zoom),
		XtRFloat,
		(XtPointer)&floatOne
	}, {
		XtNrichtextDelay,
		XtCRichtextDelay,
		XtRInt,
		sizeof(int),
		offset(delay),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNrichtextFormat,
		XtCRichtextFormat,
		XtRInt,
		sizeof(int),
		offset(format),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNrichtextString,
		XtCRichtextString,
		XtRPointer,
		sizeof(XtPointer),
		offset(rc_string),
		XtRImmediate,
		(XtPointer)NULL
	}, {
		XtNrichtextRuler,
		XtCRichtextRuler,
		XtRWidget,
		sizeof(Widget),
		offset(ruler),
		XtRImmediate,
		(XtPointer)None,
	}
};
#undef offset

/* methods */
static void DoLayout(RichtextWidget);
static void Resize(Widget);
static XtGeometryResult GeometryManager(Widget,
		XtWidgetGeometry *, XtWidgetGeometry *);
static void ChangeManaged(Widget);
static void Redisplay(Widget, XEvent *, Region);
static void Initialize(Widget, Widget, ArgList, Cardinal *);
static void Realize(Widget, XtValueMask *, XSetWindowAttributes *);
static void Destroy(Widget);
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal *);

static int default_row_height(XtPointer p, int i)
{
	RichtextWidget rtw = (RichtextWidget)p;

	if (rtw->richtext.rc_string == NULL || rtw->richtext.rc_string[0].c == '\0')
		return 20;
	return rc_strheight(rtw->richtext.rc_string, -1);
}

static int default_adj_horiz(XtPointer p, int i)
{
	return HADJ_LEFT;
}

static int default_style(XtPointer p, int i)
{
	return 0;
}

static rich_char *default_text(XtPointer p, int i)
{
	RichtextWidget rw = (RichtextWidget)p;
	if (i != 1) return NULL;
	return rw->richtext.rc_string;
}

/* by default, do not display Beginning Of Paragraph marks */
static Boolean default_bop(XtPointer p, int i)
{
	return False;
}

static void plugin_coords(Widget w, XtPointer p, int *x, int *y)
{
	*x = *y = 0;
}

/* actions */
static void MoveForwardChar(Widget, XEvent *, String *, Cardinal *);
static void MoveBackwardChar(Widget, XEvent *, String *, Cardinal *);
static void MoveToLineStart(Widget, XEvent *, String *, Cardinal *);
static void MoveToLineEnd(Widget, XEvent *, String *, Cardinal *);
static void DeleteForwardChar(Widget, XEvent *, String *, Cardinal *);
static void DeleteBackwardChar(Widget, XEvent *, String *, Cardinal *);
static void InsertChar(Widget, XEvent *, String *, Cardinal *);
static void DeleteToEnd(Widget, XEvent *, String *, Cardinal *);
static void DeleteHighlighted(Widget, XEvent *, String *, Cardinal *);
static void SelectStart(Widget, XEvent *, String *, Cardinal *);
static void ExtendAdjust(Widget, XEvent *, String *, Cardinal *);
static void ExtendEnd(Widget, XEvent *, String *, Cardinal *);
static void InsertSelection(Widget, XEvent *, String *, Cardinal *);

static XtActionsRec actions[] =
{
	{"forward-character",		MoveForwardChar},
	{"backward-character",		MoveBackwardChar},
	{"beginning-of-line",		MoveToLineStart},
	{"end-of-line",			MoveToLineEnd},
	{"delete-next-character",	DeleteForwardChar},
	{"delete-previous-character",	DeleteBackwardChar},
	{"insert-char",			InsertChar},
	{"delete-to-end",		DeleteToEnd},
	{"delete-highlighted",		DeleteHighlighted},
	{"select-start",		SelectStart},
	{"extend-adjust",		ExtendAdjust},
	{"extend-end",			ExtendEnd},
	{"insert-selection",		InsertSelection},
};

/* translations */
static char translations[] =
"<Key>Home:		beginning-of-line() \n"
":<Key>KP_Home:		beginning-of-line() \n"
"Ctrl<Key>a:		beginning-of-line() \n"
"<Key>End:		end-of-line() \n"
":<Key>KP_End:		end-of-line() \n"
"Ctrl<Key>e:		end-of-line() \n"
"<Key>Right:		forward-character() \n"
":<Key>KP_Right:	forward-character() \n"
"Ctrl<Key>f:		forward-character() \n"
"<Key>Left:		backward-character() \n"
":<Key>KP_Left:		backward-character() \n"
"Ctrl<Key>b:		backward-character() \n"
"<Key>Delete:		delete-next-character() \n"
":<Key>KP_Delete:	delete-next-character() \n"
"Ctrl<Key>d:		delete-next-character() \n"
"<Key>BackSpace:	delete-previous-character() \n"
"Ctrl<Key>k:		delete-to-end() \n"
"Ctrl<Key>w:		delete-highlighted() \n"
"<Key>:			insert-char()\n"
"<Btn1Down>:		select-start() \n"
"<Btn1Motion>:		extend-adjust() \n"
"<Btn1Up>:		extend-end() \n"
"<Btn2Down>:		insert-selection() \n"
;

RichtextClassRec richtextClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &compositeClassRec,
    /* class_name		*/	"Richtext",
    /* widget_size		*/	sizeof(RichtextRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },{
/* composite_class fields */
    /* geometry_manager   */    GeometryManager,
    /* change_managed     */    ChangeManaged,
    /* insert_child       */    XtInheritInsertChild,
    /* delete_child       */    XtInheritDeleteChild,
    /* extension          */    NULL
  }, { /* richtext fields */
    /* empty			*/	0
  }
};

WidgetClass richtextWidgetClass = (WidgetClass)&richtextClassRec;


static int ret_style(RichtextWidget rtw, int row)
{
	if (rtw->richtext.style)
		return (*rtw->richtext.style)(rtw->richtext.data, row);
	return STY_DEFAULT;
}

static rich_char *ret_text(RichtextWidget rtw, int row)
{
	if (rtw->richtext.text)
		return (*rtw->richtext.text)(rtw->richtext.data, row);
	return NULL;
}

static int line_length(RichtextWidget rtw, int row)
{
	rich_char *p;
	int len;
	p = ret_text(rtw, row);
	if (p == NULL) return 0;
	if (ret_style(rtw, row) == STY_EMBED) return 1;
	len = rc_strlen(p);
	return len;
}

/* supporting code copied directly from window.c */

static void MoveForwardChar(Widget w, XEvent *event, String *params, Cardinal *n)
{
	RichtextWidget rw = (RichtextWidget)w;
	int i = rw->richtext.point_col;
	rich_char *p = rw->richtext.rc_string;

	if (p[i].c) i++;
	rw->richtext.point_col = i;
	Redisplay(w, NULL, None);
}

static void MoveBackwardChar(Widget w, XEvent *event, String *params, Cardinal *n)
{
	RichtextWidget rw = (RichtextWidget)w;
	int i = rw->richtext.point_col;

	if (i) i--;
	rw->richtext.point_col = i;
	Redisplay(w, NULL, None);
}

static void MoveToLineStart(Widget w, XEvent *event, String *params, Cardinal *n)
{
	RichtextWidget rw = (RichtextWidget)w;
	rw->richtext.point_col = 0;
	Redisplay(w, NULL, None);
}

static void MoveToLineEnd(Widget w, XEvent *event, String *params, Cardinal *n)
{
	RichtextWidget rw = (RichtextWidget)w;
	rw->richtext.point_col = rc_strlen(rw->richtext.rc_string);
	Redisplay(w, NULL, None);
}

static void DeleteForwardChar(Widget w, XEvent *event, String *params, Cardinal *n)
{
	RichtextWidget rw = (RichtextWidget)w;
	int i = rw->richtext.point_col;
	rich_char *p = rw->richtext.rc_string;
	if (!p[i].c) return;

	rc_strcpy(p+i, p+i+1);
	Redisplay(w, NULL, None);
}

static void DeleteBackwardChar(Widget w, XEvent *event, String *params, Cardinal *n)
{
	RichtextWidget rw = (RichtextWidget)w;
	int i = rw->richtext.point_col;
	rich_char *p = rw->richtext.rc_string;

	if (!i) return;

	i--;
	rc_strcpy(p+i, p+i+1);
	rw->richtext.point_col = i;
	Redisplay(w, NULL, None);
}

static void InsertChar(Widget w, XEvent *event, String *params, Cardinal *n)
{
	RichtextWidget rw = (RichtextWidget)w;
	int count, bufsiz = 100;
	unsigned char buf[120];
	rich_char *rp;
	KeySym keysym;

	count = RichtextLookupString(w, event, buf, bufsiz, &keysym);
	if (keysym >= ' ' && count == 1) {
		buf[count] = '\0';
		rp = rc_makerich(buf, rw->richtext.format);
		RichtextInsertText(w, rp, count);
		cfree(rp);
		rw->richtext.point_col += count;
		Redisplay(w, NULL, None);
	}
}

static void DeleteToEnd(Widget w, XEvent *event, String *params, Cardinal *n)
{
	RichtextWidget rw = (RichtextWidget)w;
	int i = rw->richtext.point_col;
	rich_char *p = rw->richtext.rc_string;

	p[i].c = '\0';
	Redisplay(w, NULL, None);
}

static void DeleteHighlighted(Widget w, XEvent *event, String *params, Cardinal *n)
{
	RichtextWidget rw = (RichtextWidget)w;
	rich_char *p = rw->richtext.rc_string;

	if (rw->richtext.sel_top_col < 0) return;
	rc_strcpy(p+rw->richtext.sel_top_col, p+rw->richtext.sel_bottom_col);
	rw->richtext.point_col = rw->richtext.sel_top_col;
	rw->richtext.sel_top_col = rw->richtext.sel_bottom_col = -1;

	Redisplay(w, NULL, None);
}

static void SelectStart(Widget w, XEvent *event, String *params, Cardinal *n)
{
	/*RichtextWidget rw = (RichtextWidget)w;*/

	Redisplay(w, NULL, None);
}

static void ExtendAdjust(Widget w, XEvent *event, String *params, Cardinal *n)
{
	/*RichtextWidget rw = (RichtextWidget)w;*/

	Redisplay(w, NULL, None);
}

static void ExtendEnd(Widget w, XEvent *event, String *params, Cardinal *n)
{
	/*RichtextWidget rw = (RichtextWidget)w;*/

	Redisplay(w, NULL, None);
}

static void InsertSelection(Widget w, XEvent *event, String *params, Cardinal *n)
{
	/*RichtextWidget rw = (RichtextWidget)w;*/

	Redisplay(w, NULL, None);
}

static GC get_gc(Widget w, unsigned long fg, unsigned long bg, Font font)
{
        unsigned long valuemask = 0;
        XGCValues values;
        GC gc = XCreateGC(XtDisplay(w), XtWindow(w), valuemask, &values);

        XSetForeground(XtDisplay(w), gc, fg);
        XSetBackground(XtDisplay(w), gc, bg);
        if (font != -1)
                XSetFont(XtDisplay(w), gc, font);
        return gc;
}

#define superclass (&coreClassRec)
static void Realize(Widget w, XtValueMask *valueMask,
		XSetWindowAttributes *attributes)
{
	RichtextWidget rtw = (RichtextWidget)w;
	unsigned long fg, bg, blockbg;
	XColor screen_color, exact_color;
	Display *dpy = XtDisplay(w);
	char *name, *class;
	XIMStyles *styles;
	int i;

	(*superclass->core_class.realize)(w, valueMask, attributes);
	fg = BlackPixelOfScreen(XtScreen(w));
	bg = rtw->core.background_pixel;
	XAllocNamedColor(dpy,
		DefaultColormap(dpy, DefaultScreen(dpy)),
		"grey", &screen_color, &exact_color);
	blockbg = screen_color.pixel;
	rtw->richtext.clear_gc = get_gc(w, bg, fg, -1);
	rtw->richtext.cell_gc = get_gc(w, fg, blockbg,
			-1/*get_font(0)*/);
	rtw->richtext.cursor_gc = get_gc(w, fg^bg, 0, -1);
	XSetFunction(dpy, rtw->richtext.cursor_gc, GXxor);
        XSetLineAttributes(dpy, rtw->richtext.cursor_gc,
			1, LineSolid, CapButt, JoinMiter);

#ifdef HAVE_XCREATEIC
	/* Set up input methods */
	XtGetApplicationNameAndClass(dpy, &name, &class);
	rtw->richtext.xim = XOpenIM(dpy, XtDatabase(dpy), name, class);
	XGetIMValues(rtw->richtext.xim,
		XNQueryInputStyle, &styles,
		(char *)0);
	for (i = 0; i < styles->count_styles; i++) {
		if (styles->supported_styles[i] ==
		    (XIMPreeditNothing | XIMStatusNothing)) break;
	}
	if (i == styles->count_styles) i = 0;
	rtw->richtext.xic = XCreateIC(rtw->richtext.xim,
		XNInputStyle, styles->supported_styles[i],
		XNClientWindow, XtWindow(w),
		(char *)0);
#endif
}

static void Initialize(Widget request, Widget new, ArgList args, Cardinal *n)
{
	RichtextWidget rtw = (RichtextWidget)new;

	rtw->richtext.timeout = None;
	rtw->richtext.plain = NULL;
	if (rtw->richtext.data == NULL)
		rtw->richtext.data = (XtPointer)new;
}

static void Destroy(Widget w)
{
	RichtextWidget rtw = (RichtextWidget)w;

	XFreeGC(XtDisplay(w), rtw->richtext.clear_gc);
	XFreeGC(XtDisplay(w), rtw->richtext.cell_gc);
	XFreeGC(XtDisplay(w), rtw->richtext.cursor_gc);
#ifdef HAVE_XCREATEIC
	XDestroyIC(rtw->richtext.xic);
	XCloseIM(rtw->richtext.xim);
#endif
	if (rtw->richtext.timeout)
		XtRemoveTimeOut(rtw->richtext.timeout);
}

static Dimension row_height(RichtextWidget rtw, int row)
{
	if (rtw->richtext.row_height)
		return (*rtw->richtext.row_height)(rtw->richtext.data, row);
	return 20;
}

/* ---
forget about x for now
*/

void richtext_global_coords(RichtextWidget tw,
		int row, int col, int *x, int *y)
{
	int i;

	*x = *y = 0;

	for (i = 1; i < row; i++)
		*y += row_height(tw, i);
}

/* ---
Check if we are in the block.
*/

static int inblock(RichtextWidget rtw, int r, int c)
{
	return ((r > rtw->richtext.sel_top_row ||
		 (r == rtw->richtext.sel_top_row &&
		  c >= rtw->richtext.sel_top_col)) &&
		 (r < rtw->richtext.sel_bottom_row ||
		  (r == rtw->richtext.sel_bottom_row &&
		   c <= rtw->richtext.sel_bottom_col)));
}

static int ret_hadj(RichtextWidget rtw, int row)
{
	if (rtw->richtext.adj_horiz)
		return (*rtw->richtext.adj_horiz)(rtw->richtext.data, row);
	return HADJ_LEFT;
}

static int line_width(RichtextWidget rtw, int row, int col)
{
	rich_char *line;
        char *p;
        int i, l;
        unsigned int width;

	line = ret_text(rtw, row);
	if (line == NULL) return 0;

	p = (char *)rc_makeplain(line);

        if (ret_style(rtw, row) == STY_EMBED) {
                if (col == 0) return 0;
                embed_size(p, &width, NULL);
                return width;
        }

        l = rc_strlen(line);
        if (col > l) col = l;
        width = 0;      /* accumulated width */
        for (i = 0; i < col && line[i].c; i++) {
                if (line[i].c == '\t') {
                        width += rtw->richtext.tab_distance;
                        width -= (width % rtw->richtext.tab_distance);
                } else {
			width += rc_width(line[i]);
                }
        }
        cfree(p);
        return width;
}

static int line_start(RichtextWidget rtw, int row)
{
        int x = 0;
        int w = rtw->richtext.paper_width-rtw->richtext.left_margin
		-rtw->richtext.right_margin;
        int hadj;
        hadj = ret_hadj(rtw, row);
        if (hadj == HADJ_CENTER)
                x = (w-line_width(rtw, row, line_length(rtw, row))) / 2;
        else if (hadj == HADJ_RIGHT)
                x = w-line_width(rtw, row, line_length(rtw, row));

        return x+rtw->richtext.left_margin+rtw->richtext.top_col;
}

/* ---
*/
void richtext_char2coords(RichtextWidget rtw,
		int cell_row, int cell_col,
		int *cell_x, int *cell_y)
{
	int r;
        *cell_y = 0;
        for (r = rtw->richtext.top_row; r < cell_row; r++)
                *cell_y += row_height(rtw, r);
        *cell_x = line_start(rtw, cell_row)
                        +line_width(rtw, cell_row, cell_col);
}

/* ---
*/
void richtext_coords2char(RichtextWidget rtw,
		int *cur_row, int *cur_col,
		int cur_x, int cur_y)
{
        int r, c, h;

        r = rtw->richtext.top_row;
        h = row_height(rtw, r);
        while (cur_y > h) {
                r++;
                h += row_height(rtw, r);
        }
        *cur_row = r;

        c = 0;
        cur_x -= line_start(rtw, r);
        cur_x += rtw->richtext.top_col;       /* it's in pixels, y'know */

        while (c < line_length(rtw, r) &&
                        cur_x > line_width(rtw, r, c)) {
                c++;
        }
        *cur_col = c;
}

/*
Change top_row and top_col so that point is visible.
The unit for top_row is lines of text from the start of the document.
The unit for top_col is unzoomed pixels from the left edge of the paper.
*/
static Boolean move_top(RichtextWidget rtw)
{
	Boolean pr_scr_flag = False;
	int cur_x, cur_y;
	int width, height;
	float zoom = rtw->richtext.zoom;
	int prh;

	if (rtw->richtext.point_row < rtw->richtext.top_row) {
		rtw->richtext.top_row = rtw->richtext.point_row;
		pr_scr_flag = True;
	}
	/* Figure out how big the window is */
	width = rtw->core.width/zoom;	/* unzoomed size */
	height = rtw->core.height/zoom;
	richtext_char2coords(rtw,
		rtw->richtext.point_row, rtw->richtext.point_col,
		&cur_x, &cur_y);
	/* cur_x and cur_y are in unzoomed pixels */

	if (cur_x < 0) {
		rtw->richtext.top_col -= cur_x;
		pr_scr_flag = TRUE;
	} else if (cur_x-width > 0) {
		rtw->richtext.top_col -= cur_x-width;
		pr_scr_flag = TRUE;
	}
	/* this isn't efficient, but it will work */
        while (cur_y < 0) {     /* this should never happen */
                rtw->richtext.top_row--;
                cur_y += row_height(rtw, rtw->richtext.top_row);
                pr_scr_flag = TRUE;
        }
	prh = row_height(rtw, rtw->richtext.point_row);
        while (cur_y + prh > height) {
                cur_y -= row_height(rtw, rtw->richtext.top_row);
                rtw->richtext.top_row++;
                pr_scr_flag = TRUE;
        }
	return pr_scr_flag;
}

static void toggle_cursor(RichtextWidget rtw)
{
	int cur_x, cur_y;
	float zoom = rtw->richtext.zoom;

	richtext_char2coords(rtw,
		rtw->richtext.point_row, rtw->richtext.point_col,
		&cur_x, &cur_y);
	XDrawLine(XtDisplay((Widget)rtw), XtWindow((Widget)rtw),
			rtw->richtext.cursor_gc, zoom*cur_x, zoom*cur_y,
                        zoom*cur_x,
			zoom*(row_height(rtw, rtw->richtext.point_row)+cur_y));
}

static void draw_line(RichtextWidget rtw, Drawable cell_win,
		int y_base, int row, int clr)
{
	int color_index;
        int i;
        unsigned long color;
        int height = row_height(rtw, row);
        int x_base = 0, old_x_base;
	Display *display = XtDisplay(rtw);
        rich_char *line = ret_text(rtw, row);
	int bop;
	float zoom = rtw->richtext.zoom;
	GC gc = rtw->richtext.cell_gc;

        if (clr) {
                /* the 4 is from trial and error; should be from font metrics */
                XClearArea(display, cell_win, 0, zoom*(y_base+4),
                                rtw->core.width, zoom*height, FALSE);
        }

        if (!line) return;
        x_base = line_start(rtw, row);

        if (ret_style(rtw, row) == STY_EMBED) {
                char *p = (char *)rc_makeplain(line);
                embed_draw(cell_win, x_base, y_base, p);
                cfree(p);
                return;         /* done */
        }

	/* draw bop marker */
	bop = rtw->richtext.bop(rtw->richtext.data, row);
	switch (bop) {
		XPoint points[3];
	case 0:		/* do nothing */
		break;
	default:	/* beginning of paragraph */
		points[0].x = zoom*(x_base-10);
		points[0].y = zoom*(y_base+4);
		points[1].x = zoom*7;
		points[1].y = zoom*(height/2-4);
		points[2].x = zoom*-points[1].x;
		points[2].y = zoom*points[1].y;
		XSetForeground(display, gc,
			BlackPixelOfScreen(XtScreen(rtw)));
		XDrawLines(display, cell_win, gc,
			points, 3, CoordModePrevious);
		break;
	}

        for (i = 0; line[i].c; i++) {
                int width;
                char b[2];
                b[0] = line[i].c;
                b[1] = '\0';

                color_index = format_table[line[i].fmt].fg;
                color = get_color(color_index);
                old_x_base = x_base;
                XSetForeground(display, gc, color);
                if (inblock(rtw, row, i)) {
			rc_draw(cell_win, gc,
				0, 0, x_base, y_base+height, line[i], 1,
				zoom);
                } else {
			rc_draw(cell_win, gc,
				0, 0, x_base, y_base+height, line[i], 0,
				zoom);
		}
                if (b[0] == '\t') {
			int new_x = x_base - rtw->richtext.left_margin
					- rtw->richtext.top_col;
			new_x += rtw->richtext.tab_distance;
			new_x -= new_x % rtw->richtext.tab_distance;
			x_base = new_x + rtw->richtext.left_margin
					+ rtw->richtext.top_col;
                } else {
			width = rc_width(line[i]);
                        x_base += width;
                }
                if (format_table[line[i].fmt].uline) {
                        XDrawLine(display, cell_win,
				gc,
                                zoom*old_x_base, zoom*(y_base+height+1),
                                zoom*x_base, zoom*(y_base+height+1));
                }
                if (format_table[line[i].fmt].strike) {
                        XDrawLine(display, cell_win,
				gc,
                                zoom*old_x_base, zoom*(y_base+height-4),
                                zoom*x_base, zoom*(y_base+height-4));
                }
        }
}

/* ---
*/
void rt_pr_line(Widget w, int row)
{
        Window cell_win = XtWindow(w);
        RichtextWidget rtw = (RichtextWidget)w;
        int i;
        int y_base = 0;

        for (i = rtw->richtext.top_row; i < row; i++)
                y_base += row_height(rtw, i);
        /* here i == row+1 and y_base is just right */
	if (rtw->richtext.visible_cursor)
		toggle_cursor(rtw);
        draw_line(rtw, cell_win, y_base, row, TRUE);
	if (rtw->richtext.visible_cursor)
		toggle_cursor(rtw);
}

void draw_richtext(RichtextWidget rtw, Drawable d)
{
        int i, y_base;
        unsigned int height;
	float zoom = rtw->richtext.zoom;

	height = rtw->core.height;
        y_base = 0;
        for (i = rtw->richtext.top_row;
			/*i <= w->buf->used_lines &&*/
			zoom*y_base < height; i++) {
                draw_line(rtw, d, y_base, i, FALSE);
                y_base += row_height(rtw, i);
        }
}

/* ---
Draw onto a pixmap.
*/

Pixmap richtext_pixmap(RichtextWidget rtw)
{
        unsigned int width, height, depth;
        Pixmap scribble;
        Widget w = (Widget)rtw;

	width = rtw->core.width;
	height = rtw->core.height;
	depth = rtw->core.depth;
	if (width > 2000 || height > 2000) return None;

        scribble = XCreatePixmap(XtDisplay(w), XtWindow(w),
                width, height, depth);
        XFillRectangle(XtDisplay(w), scribble, rtw->richtext.clear_gc,
                0, 0, width, height);
        draw_richtext(rtw, scribble);
        return scribble;
}

static void update_ruler(Widget w)
{
	RichtextWidget rtw = (RichtextWidget)w;

	if (rtw->richtext.ruler) {
		XtVaSetValues(rtw->richtext.ruler,
			XtNrulerTopCol, rtw->richtext.top_col,
			XtNleftMargin, rtw->richtext.left_margin,
			XtNrightMargin, rtw->richtext.right_margin,
			XtNpaperWidth, rtw->richtext.paper_width,
			XtNtabDistance, rtw->richtext.tab_distance,
			(char *)0);
		RulerSetZoom(rtw->richtext.ruler, rtw->richtext.zoom);
	}
}

static void do_redisplay(XtPointer client_data, XtIntervalId *id)
{
	Widget w = (Widget)client_data;
        Pixmap scribble;
        GC gc;
        unsigned long valuemask = 0;
        XGCValues values;
	RichtextWidget rtw = (RichtextWidget)w;

	move_top(rtw);

	scribble = richtext_pixmap(rtw);
        if (scribble == None) return;
        gc = XCreateGC(XtDisplay(w), XtWindow(w),
                        valuemask, &values);
        XCopyArea(XtDisplay(w), scribble, XtWindow(w),
                gc, 0, 0, rtw->core.width, rtw->core.height, 0, 0);
        XFreePixmap(XtDisplay(w), scribble);
        XFreeGC(XtDisplay(w), gc);
	if (rtw->richtext.visible_cursor) {
		toggle_cursor(rtw);
	}
	/* update plugin positions */
	DoLayout(rtw);
	
	/* clear the timeout */
	rtw->richtext.timeout = None;
	update_ruler(w);
}

static void Redisplay(Widget w, XEvent *event, Region r)
{
	RichtextWidget tw = (RichtextWidget)w;
	if (tw->richtext.timeout) return;	/* already set */
	if (tw->richtext.delay) {
		tw->richtext.timeout = XtAppAddTimeOut(
				XtWidgetToApplicationContext(w),
				tw->richtext.delay, do_redisplay, (XtPointer)w);
		update_ruler(w);
	} else {
		do_redisplay((XtPointer)w, NULL);
	}
}

static Boolean SetValues(Widget current, Widget request, Widget new,
		ArgList args, Cardinal *nargs)
{
	RichtextWidget currtw = (RichtextWidget)current;
	RichtextWidget newrtw = (RichtextWidget)new;
	Boolean do_redisplay = False;

	do_redisplay = (currtw->richtext.sel_top_row != newrtw->richtext.sel_top_row
		|| currtw->richtext.sel_top_col != newrtw->richtext.sel_top_col
		|| currtw->richtext.sel_bottom_row != newrtw->richtext.sel_bottom_row
		|| currtw->richtext.sel_bottom_col != newrtw->richtext.sel_bottom_col);
	if (newrtw->richtext.visible_cursor)
		do_redisplay |= move_top(newrtw);

	if (newrtw->richtext.redisplay) {
		do_redisplay = True;
		newrtw->richtext.redisplay = False;
	}

	/* can't let Xt handle this because it will flicker */
	if (do_redisplay) {
		Redisplay((Widget)newrtw, NULL, None);
		do_redisplay = False;
	} else {
		if (currtw->richtext.visible_cursor)
			toggle_cursor(currtw);
		if (newrtw->richtext.visible_cursor)
			toggle_cursor(newrtw);
	}

	return do_redisplay;
}

/*
 * Do a layout, actually assigning positions.
 */

static void DoLayout(RichtextWidget sw)
{
        int i;
        int x, y, top_x, top_y;
	float zoom = sw->richtext.zoom;

        if (sw->composite.num_children) {
                richtext_global_coords(sw,
                        sw->richtext.top_row, sw->richtext.top_col,
                        &top_x, &top_y);
        }
        for (i = 0; i < sw->composite.num_children; i++) {
                (*sw->richtext.plugin_coords)(sw->composite.children[i],
                        sw->richtext.data, &x, &y);
		x = (sw->richtext.left_margin
			+sw->richtext.paper_width
			-sw->richtext.right_margin
			-sw->composite.children[i]->core.width)/2;
                XtMoveWidget(sw->composite.children[i],
                        zoom*(x-top_x), zoom*(y-top_y));
        }
}

/*
 * Actually layout the table
 */

static void Resize(Widget w)
{
        DoLayout((RichtextWidget)w);
}

/*
 * Geometry Manager
 */

static XtGeometryResult GeometryManager(Widget w,
                XtWidgetGeometry *request, XtWidgetGeometry *reply)
{
        return XtGeometryYes;
}

static void ChangeManaged(Widget w)
{
    DoLayout((RichtextWidget)w);
}

int RichtextLookupString(Widget w, XEvent *event, char *buf,
			 int bufsiz, KeySym *keysym)
{
#ifdef HAVE_XCREATEIC
	Status status;
	return XmbLookupString(((RichtextWidget)w)->richtext.xic,
				(XKeyEvent *)event, (char *)buf, bufsiz,
				keysym, &status);
#else
        return XLookupString((XKeyEvent *)event, (char *)buf, bufsiz,
				keysym, NULL);
#endif
}

void RichtextSetZoom(Widget w, float zoom)
{
	RichtextWidget rw = (RichtextWidget)w;

	if (zoom < .1) zoom = .1;
	if (zoom > 10) zoom = 10;
	if (rw->richtext.zoom != zoom) {
		rw->richtext.zoom = zoom;
		Redisplay(w, NULL, None);
	}
}

/*
Functions that allow this widget to replace TextField with a minimum of effort.
*/
void RichtextSetString(Widget w, char *p)
{
	RichtextWidget rw = (RichtextWidget)w;

	cfree(rw->richtext.rc_string);
	rw->richtext.rc_string = rc_makerich((unsigned char *)p,
						rw->richtext.format);
	Redisplay(w, NULL, None);
}

char *RichtextGetString(Widget w)
{
	RichtextWidget rw = (RichtextWidget)w;
	cfree(rw->richtext.plain);
	rw->richtext.plain = (char *)rc_makeplain(rw->richtext.rc_string);
	return rw->richtext.plain;
}

void RichtextInsertText(Widget w, rich_char *buf, int count)
{
	RichtextWidget rw = (RichtextWidget)w;
	int i = rw->richtext.point_col;
	rich_char *p = rw->richtext.rc_string;
	int n = rc_strlen(p);
	rich_char *q = cmalloc((n+count+1)*sizeof *q);
	rc_strncpy(q, p, i);
	rc_strncpy(q+i, buf, count);
	rc_strcpy(q+i+count, p+i);
	rw->richtext.rc_string = q;
	cfree(p);
}

